import config from '../config'
import randomItem from '../util/randomItem'
import removeRandomItem from '../util/removeRandomItem'
import StorageDisplay from '../widgets/StorageDisplay'
import TimeDisplay from '../widgets/TimeDisplay'
import NextCustomerDialog from '../widgets/NextCustomerDialog'
import RepairMaterialSelectionDisplay from '../widgets/RepairMaterialSelectionDisplay'
import CustomerDisplay from '../widgets/CustomerDisplay'
import ToolsSelection from '../widgets/ToolsSelection'
import ToolsDropper from '../widgets/ToolsDropper'
import LootDialog from '../widgets/LootDialog'

export default class MainScene extends Phaser.Scene {
	constructor () {
		super({ key: 'Main' })
		
		this.backgroundMusic = null
	}

	create () {
		if (this.backgroundMusic === null) {
			this.backgroundMusic = this.sound.add('theme', {
				loop: true
			})
			this.backgroundMusic.setVolume(0.25)
			this.backgroundMusic.play()
		}
		
		this.dropSound = this.sound.add('drop')
		this.dropSound.setVolume(0.25)
		this.clickSound = this.sound.add('click')
		this.clickSound.setVolume(0.25)
		this.lootSound = this.sound.add('loot')
		this.lootSound.setVolume(0.25)
		
		this.currentDay = 1
		this.timeOfDay = 8 * 60
		this.waitingCustomers = this.generateWaitingCustomers(this.currentDay, config.armors)
		this.storage = this.createInitialStorage(config.initiallyStoredMaterials, config.materials)
		this.availableStorageItems = Object.keys(this.storage).reduce((accumulator, materialId) => {
			for (let i = 0; i < this.storage[materialId]; i++) {
				accumulator.push(materialId)
			}
			return accumulator
		}, [])
		this.selectedMaterials = []
		this.doneCustomers = []
		this.timeTween = null
		
		this.background = new Phaser.GameObjects.Graphics(this)
		this.background.lineStyle(10, 0x7f401c, 1.0)
		this.background.fillStyle(0xc67c33, 1.0)
		this.background.fillRect(0, 0, this.sys.canvas.width, 180)
		this.background.beginPath()
		this.background.moveTo(0, 180)
		this.background.lineTo(this.sys.canvas.width, 180)
		this.background.strokePath()
		
		this.storageDisplay = new StorageDisplay({
			scene: this,
			background: true
		})
		this.storageDisplay.updateCounts(this.storage)
		this.storageDisplay.setPosition(0, this.sys.canvas.height - 450)
		
		this.timeDisplay = new TimeDisplay({
			scene: this
		})
		this.timeDisplay.setPosition(0, 50)
		this.timeDisplay.updateTimeOfDay(this.timeOfDay)
		
		this.dayText = new Phaser.GameObjects.Text(this, 50, 50, 'Day 1', config.timeTextStyle)
		
		this.nextCustomerDialog = new NextCustomerDialog({
			scene: this,
			onClose: () => {
				this.startRepair()
			}
		})
		this.nextCustomerDialog.setPosition(this.sys.canvas.width / 2, this.sys.canvas.height / 2 - 200)
		
		this.repairMaterialSelectionDisplay = new RepairMaterialSelectionDisplay({
			scene: this,
			onDropped: (index) => {
				this.updateMaterial(index)
				this.dropSound.play()
			}
		})
		this.repairMaterialSelectionDisplay.visible = false
		this.repairMaterialSelectionDisplay.setPosition(0, this.sys.canvas.height - 300)
		
		this.currentCustomerDisplay = new CustomerDisplay({
			scene: this,
			customer: {
				armorId: '',
				holes: []
			},
			onArmorFixed: () => {
				this.startToolchain()
			}
		})
		this.currentCustomerDisplay.visible = false
		this.currentCustomerDisplay.setPosition(this.sys.canvas.width / 2, this.sys.canvas.height / 2)
		
		this.toolsSelection = new ToolsSelection({
			scene: this,
			onToolSelected: (toolId) => {
				this.toolsDropper.tryMatch(toolId)
				this.clickSound.play()
			}
		})
		this.toolsSelection.visible = false
		this.toolsSelection.setPosition(0, this.sys.canvas.height - 300)
		
		this.toolsDropper = new ToolsDropper({
			scene: this,
			dropLength: this.sys.canvas.height * 0.65,
			onSuccess: () => {
				this.finishRepair()
			},
			onFail: () => {
				this.gameOver()
			}
		})
		this.toolsDropper.visible = false
		this.toolsDropper.setPosition(150, 200)
		
		this.lootDialog = new LootDialog({
			scene: this,
			onClose: () => {
				this.currentDay++
				this.startNewDay()
			}
		})
		this.lootDialog.setPosition(0, 300)
		
		this.dragTutorialText = new Phaser.GameObjects.Text(this, this.sys.canvas.width / 2, this.sys.canvas.height - 550, 'Drag to repair', config.dialogTextStyle)
		this.dragTutorialText.setOrigin(0.5, 0)
		this.dragTutorialText.visible = false
		
		this.matchTutorialText = new Phaser.GameObjects.Text(this, this.sys.canvas.width / 2, this.sys.canvas.height - 550, 'Match it!', config.dialogTextStyle)
		this.matchTutorialText.setOrigin(0.5, 0)
		this.matchTutorialText.visible = false
		
		this.add.existing(this.background)
		this.add.existing(this.storageDisplay)
		this.add.existing(this.timeDisplay)
		this.add.existing(this.dayText)
		this.add.existing(this.nextCustomerDialog)
		this.add.existing(this.currentCustomerDisplay)
		this.add.existing(this.dragTutorialText)
		this.add.existing(this.repairMaterialSelectionDisplay)
		this.add.existing(this.matchTutorialText)
		this.add.existing(this.toolsSelection)
		this.add.existing(this.toolsDropper)
		this.add.existing(this.lootDialog)
		// 1. Werkstatt
		// - Kundenwarteschlange (Länge nach Schwierigkeit = Tag)
		// - Kunde anzeigen
		// - Tagezeit (Uhr + Leiste)
		// - Lager
		// Auftrag -> Rüstung mit Loch
		// repariert
		// Materialien aus Lager holen
		// Lager
		// Werkzeuge + Dance Dance Revolution
		// Tageszeit
		// Nacht + Loot
		// Gameover
		
		this.nextCustomerDialog.show(this.waitingCustomers.length)
	}
	
	createInitialStorage (count, materials) {
		const storage = materials.reduce((accumulator, material) => {
			accumulator[material.id] = 0
			return accumulator
		}, {})
		
		for (let i = 0; i < count; i++) {
			const randomMaterial = randomItem(materials)
			
			storage[randomMaterial.id]++
		}
		
		return storage
	}
	
	generateWaitingCustomers (difficulty, availableArmors) {
		const waitingCustomers = []
		const customerCount = Math.sqrt(difficulty) * 2 + 1
		for (let i = 0; i < customerCount; i++) {
			waitingCustomers.push(this.randomCustomer(difficulty, availableArmors))
		}
		return waitingCustomers
	}
	
	randomCustomer (difficulty, availableArmors) {
		const customer = {}
		const randomArmor = randomItem(availableArmors)
		customer.armorId = randomArmor.id
		customer.holes = [randomItem(randomArmor.possibleHoles)]
		customer.fixes = [undefined]
		return customer
	}
	
	updateMaterial (index) {
		this.storage[this.selectedMaterials[index]]--
		this.selectedMaterials[index] = undefined
		if (this.availableStorageItems.length > 0) {
			this.selectedMaterials[index] = removeRandomItem(this.availableStorageItems)
		}
		this.repairMaterialSelectionDisplay.updateMaterials(this.selectedMaterials)
	}
	
	startRepair () {
		for (let i = 0; i < config.numberOfSelectedMaterials && this.availableStorageItems.length > 0; i++) {
			this.selectedMaterials.push(removeRandomItem(this.availableStorageItems))
		}
		this.repairMaterialSelectionDisplay.updateMaterials(this.selectedMaterials)
		this.repairMaterialSelectionDisplay.visible = true
		this.storageDisplay.visible = false
		
		this.currentCustomer = this.waitingCustomers.pop()
		this.currentCustomerDisplay.updateCustomer(this.currentCustomer)
		this.currentCustomerDisplay.visible = true
		this.dragTutorialText.visible = true
		
		if (!this.timeTween) {
			this.timeTween = this.tweens.add({
				targets: this,
				timeOfDay: 18 * 60,
				ease: 'Linear',
				duration: 10 * 5 * 1000,
				onUpdate: () => {
					this.timeDisplay.updateTimeOfDay(this.timeOfDay)
				},
				onComplete: () => {
					if (this.waitingCustomers.length === 0) {
						this.finishDay()
					} else {
						this.gameOver()
					}
					this.timeTween = null
				}
			})
		}
	}
	
	generateToolchain (difficulty) {
		const toolchain = []
		const toolsCount = Math.sqrt(difficulty) * 2 + 1
		const duration = 1500 + 1500 / difficulty
		let delay = 0
		for (let i = 0; i < toolsCount; i++) {
			toolchain.push({
				tool: randomItem(config.tools),
				duration,
				delay
			})
			delay += 500 + Math.floor(Math.random() * 1000 / difficulty)
		}
		return toolchain
	}
	
	startToolchain () {
		this.repairMaterialSelectionDisplay.visible = false
		this.dragTutorialText.visible = false
		this.availableStorageItems = this.availableStorageItems.concat(this.selectedMaterials.filter((material) => material !== undefined))
		this.selectedMaterials = []
		
		const toolchain = this.generateToolchain(this.currentDay)
		
		this.toolsDropper.updateToolchain(toolchain)
		this.toolsDropper.visible = true
		this.toolsSelection.visible = true
		this.matchTutorialText.visible = true
	}
	
	gameOver () {
		this.scene.start('GameOver', { days: this.currentDay })
	}
	
	finishRepair () {
		this.doneCustomers.push(this.currentCustomer)
		
		this.toolsDropper.visible = false
		this.toolsSelection.visible = false
		this.currentCustomerDisplay.visible = false
		this.dragTutorialText.visible = false
		this.matchTutorialText.visible = false
		
		if (this.waitingCustomers.length === 0 || this.availableStorageItems.length === 0) {
			this.timeTween.timeScale = 25
		} else {
			this.nextCustomerDialog.show(this.waitingCustomers.length)
		}
		
		this.storageDisplay.updateCounts(this.storage)
		this.storageDisplay.visible = true
	}
	
	finishDay () {
		const loot = config.materials.reduce((accumulator, material) => {
			accumulator[material.id] = 0
			return accumulator
		}, {})
		let surviveCounter = 0
		
		for (const doneCustomer of this.doneCustomers) {
			const armorSpec = config.armors.find((armor) => armor.id === doneCustomer.armorId) || {}
			let stability = 0
			let matchesBaseMaterial = true
			for (const fix of doneCustomer.fixes) {
				const material = config.materials.find((material) => material.id === fix)
				stability += material.stability
				matchesBaseMaterial = matchesBaseMaterial && material.id === armorSpec.materialId
			}
			stability /= doneCustomer.fixes.length
			if (matchesBaseMaterial) {
				stability += 5
			}
			
			const chance = Math.random() * 8
			const survived = chance <= stability
			
			if (survived) {
				surviveCounter++
				for (let i = 0; i < 2; i++) {
					const lootedMaterial = randomItem(config.materials).id
					loot[lootedMaterial]++
					this.storage[lootedMaterial]++
					this.availableStorageItems.push(lootedMaterial)
				}
			}
		}
		
		this.repairMaterialSelectionDisplay.visible = false
		this.storageDisplay.visible = false
		this.toolsDropper.visible = false
		this.toolsSelection.visible = false
		this.currentCustomerDisplay.visible = false
		this.dragTutorialText.visible = false
		this.matchTutorialText.visible = false

		this.lootDialog.show(this.doneCustomers.length, surviveCounter, loot)
		this.lootSound.play()
		this.storageDisplay.updateCounts(this.storage)
	}
	
	startNewDay () {
		this.repairMaterialSelectionDisplay.visible = false
		this.storageDisplay.visible = true
		this.toolsDropper.visible = false
		this.toolsSelection.visible = false
		this.currentCustomerDisplay.visible = false
		this.dragTutorialText.visible = false
		this.matchTutorialText.visible = false
		
		this.doneCustomers = []
		this.waitingCustomers = this.generateWaitingCustomers(this.currentDay, config.armors)
		this.nextCustomerDialog.show(this.waitingCustomers.length)
		this.timeOfDay = 8 * 60
		this.timeDisplay.updateTimeOfDay(this.timeOfDay)
		this.dayText.setText('Day ' + this.currentDay)
	}
}
